@@ -14,11 +14,15 @@
 /** \cond */
 #include <stdbool.h>
 #include <glib-2.0/glib.h>
+
+#include "error_info.h"
 /** \endcond */
 
 
 // Issue:  Where to send error messages?
 
+extern bool failsim_enabled;
+
 //
 // Initialization
 //
@@ -29,7 +33,7 @@ void fsim_set_name_to_number_funcs(
       Fsim_Name_To_Number_Func  func,
       Fsim_Name_To_Number_Func  unmodulated_func);
 
-/** Indicates whether a failure should occur exactly one or be recurring */
+/** Indicates whether a failure should occur exactly once or be recurring */
 typedef enum {FSIM_CALL_OCC_RECURRING, FSIM_CALL_OCC_SINGLE} Fsim_Call_Occ_Type;
 
 
@@ -44,7 +48,7 @@ void fsim_add_error(
        int                  rc);
 void fsim_clear_errors_for_func(char * funcname);
 void fsim_clear_error_table();
-void fsim_report_error_table(int depth);
+void fsim_report_failure_simulation_table(int depth);
 void fsim_reset_callct(char * funcname);
 
 
@@ -52,10 +56,8 @@ void fsim_reset_callct(char * funcname);
 // Bulk load error table
 //
 
-bool fsim_load_control_from_gptrarray(GPtrArray * lines);
-
+bool fsim_load_control_from_gptrarray(GPtrArray * lines, GPtrArray * err_msg_accumulator);
 // bool fsim_load_control_string(char * s);         // unimplemented
-
 bool fsim_load_control_file(char * fn);
 
 
@@ -76,8 +78,8 @@ typedef struct {
 
 Failsim_Result fsim_check_failure(const char * fn, const char * funcname);
 
+#ifdef UNUSED
 #ifdef ENABLE_FAILSIM
-
 #define FAILSIM \
    do { \
       Failsim_Result __rcsim = fsim_check_failure(__FILE__, __func__); \
@@ -85,7 +87,6 @@ Failsim_Result fsim_check_failure(const char * fn, const char * funcname);
          return __rcsim.failure_value;  \
    } while(0);
 
-
 #define FAILSIM_EXT(__addl_cmds) \
    do { \
       Failsim_Result __rcsim = fsim_check_failure(__FILE__, __func__); \
@@ -94,13 +95,26 @@ Failsim_Result fsim_check_failure(const char * fn, const char * funcname);
          return __rcsim.failure_value;  \
       } \
    } while(0);
-
 #else
-
 #define FAILSIM
-
 #define FAILSIM_EXT(__addl_cmds)
+#endif
+#endif
 
+bool        fsim_bool_injector(bool status, const char * fn, const char * func);
+int         fsim_int_injector(int status, const char * fn, const char * func);
+Error_Info* fsim_errinfo_injector(Error_Info* status, const char * fn, const char * func);
+
+#ifdef UNUSED
+#ifdef ENABLE_FAILSIM
+#define INJECT_BOOL_ERROR(status)    fsim_bool_injector(status, __FILE__, __func__)
+#define INJECT_INT_ERROR(status)     fsim_int_injector(status, __FILE__, __func__)
+#define INJECT_ERRINFO_ERROR(status) fsim_errinfo_injector(status, __FILE__, __func__)
+#else
+#define INJECT_BOOL_ERROR(status)  status
+#define INJECT_INT_ERROR(status)   status
+#define INJECT_ERRINFO_ERROR(status) status
+#endif
 #endif
 
 #endif /* FAILSIM_H_ */
